module Parse(Parser,parses,parse,token,tokens,(***),(+++),
             always,never,star,plus,char,guard,int) where

import Char
import Test.QuickCheck
import List(nub,sort)
import Control.Monad(liftM)

-- Precedence declarations
-- *** is left associative, so
-- p *** q *** r  means  (p *** q) *** r

-- *** has higher precedence than +++, so
-- p +++ q *** r  means  p +++ (q *** r)

infixl 3 ***
infixr 2 +++
infix  1 ===,====


-- The type of parsers

newtype  Parser a = Make (String -> [(a, String)])


-- Apply parser to input

apply :: Parser a -> String -> [(a,String)]
apply (Make p) s  =  [ (x,t) | (x,t) <- p s ]


-- Find all parses of input

parses :: Parser a -> String -> [a]
parses p s  =  [ x | (x,t) <- apply p s, t == "" ]


-- Find unique parse of input

parse :: Parser a -> String -> a
parse p s  =  extract (parses p s)
  where
  extract []        =  error "no parse"
  extract [x]       =  x
  extract (x:y:zs)  =  error "ambiguous parses"


-- Create a parser that maches a given character

token c  =  Make r
  where
  r []                 =  []
  r (d:s) | c == d     =  [(d,s)]
          | otherwise  =  []


-- Parse a single item

char :: Parser Char
char = Make r
  where
  r []     =  []
  r (c:s)  =  [(c,s)]


-- Parse a given string

tokens :: String -> Parser String
tokens []     =  always []
tokens (c:s)  =  always (:) *** token c *** tokens s



-- Find parses satisfying the given predicate

guard :: (a -> Bool) -> Parser a -> Parser a
guard p (Make q)  =  Make r
  where
  r s  =  [ (x,t) | (x,t) <- q s, p x ]


-- alternation combinator

(+++) :: Parser a -> Parser a -> Parser a
Make p +++ Make q  =  Make r
  where
  r s = p s ++ q s


-- sequential combinator

(***) :: Parser (a -> b) -> Parser a -> Parser b
Make p *** Make q  =  Make r
  where
  r s = [(f x, u) | (f, t) <- p s, (x, u) <- q t ]


-- the parser that always succeeds

always :: a -> Parser a
always x  =  Make r
  where
  r s  =  [(x,s)]


-- the parser that never succeeds

never :: Parser a
never  =  Make r
  where
  r s  =  []


-- parse zero or more occurrences

star :: Parser a -> Parser [a]
star p  =  always (:) *** p *** star p +++ always []


-- parse one or more occurrences

plus :: Parser a -> Parser [a]
plus p  =  always (:) *** p *** star p


-- example parsers

digit, lower, vowel :: Parser Char
digit = guard isDigit char                 -- parse a digit
lower = guard isLower char                 -- parse a lower case letter
vowel = guard (`elem` "aAeEiIoOuU") char   -- parse any letter

dorv :: Parser Char
dorv =  lower +++ vowel                   -- parse lower or vowel

dandv :: Parser String                    -- parse lower followed by vowel
dandv = always pair *** lower *** vowel
  where
  pair x y  =  [x,y]

paren :: Parser a -> Parser a           -- parse p surrounded by parentheses
paren p  = always middle *** token '(' *** p *** token ')'
  where
  middle '(' x ')'  =  x

posint :: Parser Int                    -- parse a positive integer (or zero)
posint = always digitsToInt *** plus digit
  where
  digitsToInt s  =  loop 0 s
  loop x []      =  x
  loop x (d:s)   =  loop (10 * x + digitToInt d) s

negint :: Parser Int                    -- parse a negative integer (or zero)
negint = always neg *** token '-' *** posint
  where
  neg '-' x  =  -x

int :: Parser Int                       -- parse an integer
int = posint +++ negint


-- Properties

prop_token :: Char -> Property
prop_token c  =  token c  ===  guard (== c) char

prop_always :: (Char -> Char) -> Char -> Property
prop_always f x   =  always f *** always x === always (f x)

prop_seq_id :: Parser Int -> Property
prop_seq_id p  =  always id *** p === p

prop_alt_id :: Parser Int -> Property
prop_alt_id p  =  never +++ p === p

prop_alt_assoc :: Parser Int -> Parser Int -> Parser Int -> Property
prop_alt_assoc p q r  =  p +++ (q +++ r) === (p +++ q) +++ r

prop_seq_assoc :: Parser (Char -> Bool -> Int) -> Parser Char -> Parser Bool -> Property
prop_seq_assoc p q r  =  p *** q *** r  ===
                         (always uncurry *** p) *** (always (,) *** q *** r)

prop_alt_comm :: Parser Int -> Parser Int -> Property
prop_alt_comm p q  =  p +++ q ==== q +++ p

prop_seq_comm :: Parser Int -> Parser Char -> Property
prop_seq_comm p q  =  always pair *** p *** q ==== always swap *** q *** p
  where
  pair x y = (x,y)
  swap x y = (y,x)

prop_distribute :: Parser (Char -> Int) -> Parser Char -> Parser Char -> Property
prop_distribute p q r  =  p *** (q +++ r) ==== (p *** q) +++ (p *** r)

instance Show (Parser a) where
  show p = "parser"

instance Show (a -> b) where
  show f = "function"

(===) :: (Show a, Eq a) => Parser a -> Parser a -> Property
p === q  =  forAll arbitrary (\ s -> not (null (apply p s)) ==>
                                     apply p s == apply q s)

(====) :: Ord a => Parser a -> Parser a -> Property
p ==== q  =  forAll arbitrary (\ s -> not (null (apply p s)) ==>
                                      nub (sort (apply p s)) == nub (sort (apply q s)))

parser :: [(Char,a)] -> Parser a
parser x  =  foldr (+++) never [ always (const v) *** token c | (c,v) <- x ]

instance Arbitrary a => Arbitrary (Parser a) where
  arbitrary = liftM parser arbitrary

main =
  quickCheck prop_token >>
  quickCheck prop_always >>
  quickCheck prop_seq_id >>
  quickCheck prop_alt_id >>
  quickCheck prop_alt_assoc >>
  quickCheck prop_seq_assoc >>
  quickCheck prop_alt_comm >>
  quickCheck prop_seq_comm >>
  quickCheck prop_distribute



  
